%{

/*
 * CalcParser.y
 *
 * Created on 30 de octubre de 2006, 22:29
 *
 * This program is the legal property of its developers, whose names are
 * listed with this program.  Please refer to the COPYRIGHT file distributed
 * with this source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

import icalc.calculation.*;
import icalc.calculation.value.*;
import icalc.calculation.Runtime;
import icalc.calculation.Package;
import icalc.stmt.*;
import icalc.stmt.exp.*;
import icalc.userops.*;
import icalc.util.*;

import javax.swing.*;
import java.util.*;
import java.io.*;
import java.lang.reflect.*;

%}

%token ASSIGN
%token OR
%token EQ
%token NE
%token LT
%token GT
%token LE
%token GE
%token AND
%token BOR
%token BXOR
%token BAND
%token LSHIFT
%token RSHIFT
%token SUB
%token ADD
%token INC
%token DEC
%token MULT
%token DIV
%token MOD
%token POW
%token NEG
%token POS
%token BNEG
%token NOT

%token ORASSIGN
%token ANDASSIGN
%token BORASSIGN
%token BXORASSIGN
%token BANDASSIGN
%token LSHIFTASSIGN
%token RSHIFTASSIGN
%token ADDASSIGN
%token SUBASSIGN
%token MULTASSIGN
%token DIVASSIGN
%token MODASSIGN
%token POWASSIGN

%token IDENTIFIER
%token COMMA
%token COLON
%token IF
%token ELSE
%token DO
%token WHILE
%token FOR
%token FUNCTION
%token RETURN
%token BREAK
%token DEF
%token INCLUDE

%token PARENT
%token PARENT_END
%token BRACE
%token BRACE_END
%token BRACKET
%token BRACKET_END

%token ENDOFINPUT

%token <sval> HELPCMD
%token <sval> EXITCMD
%token <sval> LOADCLASSCMD
%token <sval> LOADJARCMD
%token <sval> CLEARCMD

%token <ival> VALUE
%token <sval> IDENTIFIER
%type  <ival> vector

%type  <ival> exp
%type  <ival> constant
%type  <ival> assignment
%type  <ival> invocation

%type  <ival> stmt
%type  <ival> open_stm
%type  <ival> clos_stm
%type  <ival> sing_stm
%type  <ival> mult_stm

%left  ASSIGN ORASSIGN ANDASSIGN BORASSIGN BXORASSIGN BANDASSIGN LSHIFTASSIGN RSHIFTASSIGN ADDASSIGN SUBASSIGN MULTASSIGN DIVASSIGN MODASSIGN POWASSIGN
%left  OR
%left  AND
%left  BOR
%left  BXOR
%left  BAND
%left  EQ NE
%left  LT GT LE GE
%left  LSHIFT RSHIFT
%left  SUB ADD
%left  MULT DIV MOD
%right POW
%right DOT
%left  NEG POS INC DEC BNEG NOT DEF INCLUDE

%%

input:    exp         ENDOFINPUT { compiledExpr = exprStack.get($1);           if (true) return 0; }
        | mult_stm_in ENDOFINPUT { compiledCode = multiStmtStack.removeLast(); if (true) return 0; }
        | command     ENDOFINPUT {                                             if (true) return 0; }
        | ENDOFINPUT             { compiledCode = new MultiStmt();             if (true) return 0; }
        ;

command:  HELPCMD               { helpOperation     .execute($1); }
        | EXITCMD               { exitOperation     .execute($1); }
        | CLEARCMD              { clearOperation    .execute($1); }
        | LOADCLASSCMD          { loadClassOperation.execute($1); }
        | LOADJARCMD            { loadJarOperation  .execute($1); }
        ;

exp:      constant              { $$ = $1;                         if ($$==-1) return -1; }
        | invocation            { $$ = $1;                         if ($$==-1) return -1; }
        | assignment            { $$ = $1;                         if ($$==-1) return -1; }
        | PARENT exp PARENT_END { $$ = $2;                         if ($$==-1) return -1; }
        | DEF exp               { $$ = operation (DEF,    $2);     if ($$==-1) return -1; }
        | NOT  exp              { $$ = operation (NOT,    $2);     if ($$==-1) return -1; }
        | BNEG exp              { $$ = operation (BNEG,   $2);     if ($$==-1) return -1; }
        | SUB exp  %prec NEG    { $$ = operation (NEG,    $2);     if ($$==-1) return -1; }
        | ADD exp  %prec POS    { $$ = operation (POS,    $2);     if ($$==-1) return -1; }
        | exp POW  exp          { $$ = operation (POW,    $1, $3); if ($$==-1) return -1; }
        | exp MULT exp          { $$ = operation (MULT,   $1, $3); if ($$==-1) return -1; }
        | exp DIV  exp          { $$ = operation (DIV,    $1, $3); if ($$==-1) return -1; }
        | exp MOD  exp          { $$ = operation (MOD,    $1, $3); if ($$==-1) return -1; }
        | exp ADD  exp          { $$ = operation (ADD,    $1, $3); if ($$==-1) return -1; }
        | exp SUB  exp          { $$ = operation (SUB,    $1, $3); if ($$==-1) return -1; }
        | exp LSHIFT exp        { $$ = operation (LSHIFT, $1, $3); if ($$==-1) return -1; }
        | exp RSHIFT exp        { $$ = operation (RSHIFT, $1, $3); if ($$==-1) return -1; }
        | exp LE exp            { $$ = operation (LE,     $1, $3); if ($$==-1) return -1; }
        | exp GE exp            { $$ = operation (GE,     $1, $3); if ($$==-1) return -1; }
        | exp LT exp            { $$ = operation (LT,     $1, $3); if ($$==-1) return -1; }
        | exp GT exp            { $$ = operation (GT,     $1, $3); if ($$==-1) return -1; }
        | exp EQ exp            { $$ = operation (EQ,     $1, $3); if ($$==-1) return -1; }
        | exp NE exp            { $$ = operation (NE,     $1, $3); if ($$==-1) return -1; }
        | exp BAND exp          { $$ = operation (BAND,   $1, $3); if ($$==-1) return -1; }
        | exp BXOR exp          { $$ = operation (BXOR,   $1, $3); if ($$==-1) return -1; }
        | exp BOR  exp          { $$ = operation (BOR,    $1, $3); if ($$==-1) return -1; }
        | exp AND  exp          { $$ = operation (AND,    $1, $3); if ($$==-1) return -1; }
        | exp OR   exp          { $$ = operation (OR,     $1, $3); if ($$==-1) return -1; }
        | INCLUDE exp           { $$ = pushExpression(new IncludeExp(lexer.getCurrentPosition(), exprStack.get($2))); }
        ;

stmt:     open_stm  { $$ = $1; }
        | clos_stm  { $$ = $1; }
        ;

open_stm: IF PARENT exp PARENT_END stmt                             { $$ = pushStmt(new ConditionalStmt(exprStack.get($3), stmtStack.get($5))); }
        | IF PARENT exp PARENT_END clos_stm ELSE open_stm           { $$ = pushStmt(new ConditionalStmt(exprStack.get($3), stmtStack.get($5), stmtStack.get($7))); }
        | WHILE PARENT exp PARENT_END open_stm                      { $$ = pushStmt(new IteratorStmt(exprStack.get($3), stmtStack.get($5), true)); }
        | FOR PARENT exp COLON exp COLON exp PARENT_END open_stm    { $$ = pushStmt(new IteratorStmt(exprStack.get($3), exprStack.get($5), exprStack.get($7), stmtStack.get($9))); }
        ;
        
clos_stm: sing_stm                                                  { $$ = $1; }
        | IF PARENT exp PARENT_END clos_stm ELSE clos_stm           { $$ = pushStmt(new ConditionalStmt(exprStack.get($3), stmtStack.get($5), stmtStack.get($7))); }
        | WHILE PARENT exp PARENT_END clos_stm                      { $$ = pushStmt(new IteratorStmt(exprStack.get($3), stmtStack.get($5), true)); }
        | WHILE PARENT exp COLON exp COLON exp PARENT_END clos_stm  { $$ = pushStmt(new IteratorStmt(exprStack.get($3), stmtStack.get($5), true)); }
        | DO BRACE stmt BRACE_END WHILE PARENT exp PARENT_END       { $$ = pushStmt(new IteratorStmt(exprStack.get($7), stmtStack.get($3), false)); }
        | FOR PARENT exp COLON exp COLON exp PARENT_END clos_stm    { $$ = pushStmt(new IteratorStmt(exprStack.get($3), exprStack.get($5), exprStack.get($7), stmtStack.get($9))); }
        | FUNCTION IDENTIFIER PARENT PARENT_END mult_stm                               { $$ = functionDeclaration($2, new Vector<String>(), $5); }
        | FUNCTION IDENTIFIER PARENT begin_param_names param_names PARENT_END mult_stm { $$ = functionDeclaration($2, paramNames.pop(), $7); }
        ;

sing_stm: exp COLON         { $$ = pushStmt(exprStack.get($1)); }
        | RETURN exp COLON  { $$ = pushStmt(new ReturnStmt(lexer.getCurrentPosition(), exprStack.get($2))); }
        | RETURN COLON      { $$ = pushStmt(new ReturnStmt(lexer.getCurrentPosition(), null)); }
        | BREAK COLON       { $$ = pushStmt(new BreakStmt()); }
        | mult_stm          { $$ = $1; }
        ;

mult_stm: BRACE BRACE_END                            { $$ = pushStmt(new MultiStmt()); }
        | BRACE mult_stm_begin mult_stm_in BRACE_END { $$ = pushStmt(multiStmtStack.removeLast()); }
        ;

mult_stm_begin:                                      { multiStmtStack.addLast(new MultiStmt()); }
        ;

mult_stm_in: mult_stm_in stmt { multiStmtStack.getLast().addStmt(stmtStack.get($2)); }
        |  stmt               { multiStmtStack.getLast().addStmt(stmtStack.get($1)); }
        ;

begin_params:                 { params.push(new Vector<Integer>()); }
        ;

params:   exp                 { params.peek().add($1); }
        | params COMMA exp    { params.peek().add($3); }
        ;

begin_param_names:            { paramNames.push(new Vector<String>()); }
        ;

param_names: IDENTIFIER                 { paramNames.peek().add($1); }
        | param_names COMMA IDENTIFIER  { paramNames.peek().add($3); }
        ;

constant: VALUE                   { $$ = readOperation ($1); }
        | IDENTIFIER              { $$ = readOperation ($1); }
        | vector                  { $$ = readOperation ($1); }
        | IDENTIFIER vector_index { $$ = readOperation ($1, vectorIndex.pop()); }
        ;

vector:   BRACKET BRACKET_END                         { $$ = pushValue(new Vector<Integer>()); }
        | BRACKET begin_vector vector_in BRACKET_END  { $$ = pushValue(vector.pop());          }
        ;

begin_vector:                   { vector.push(new Vector<Integer>()); }
        ;

vector_in: exp                  { vector.peek().add($1); }
        | vector_in COMMA exp   { vector.peek().add($3); }
        ;

invocation: IDENTIFIER PARENT PARENT_END                            { $$ = functionInvocation ($1, null);             }
        | IDENTIFIER PARENT begin_params params PARENT_END          { $$ = functionInvocation ($1, params.pop());     }
        | exp DOT IDENTIFIER PARENT begin_params params PARENT_END  { $$ = functionInvocation ($3, $1, params.pop()); }
        ;

assignment: IDENTIFIER ASSIGN                       { $$ = createAssignment($1);      }
        | IDENTIFIER ASSIGN exp                     { $$ = createAssignment($1, $3);  }
        | IDENTIFIER INC                            { $$ = createOperationAndAssignment(ADD,    true,  $1, new IntegerValue(1)); }
        | IDENTIFIER DEC                            { $$ = createOperationAndAssignment(SUB,    true,  $1, new IntegerValue(1)); }
        | INC IDENTIFIER                            { $$ = createOperationAndAssignment(ADD,    false, $2, new IntegerValue(1)); }
        | DEC IDENTIFIER                            { $$ = createOperationAndAssignment(SUB,    false, $2, new IntegerValue(1)); }
        | IDENTIFIER ORASSIGN exp                   { $$ = createOperationAndAssignment(OR,     false, $1, $3); }
        | IDENTIFIER ANDASSIGN exp                  { $$ = createOperationAndAssignment(AND,    false, $1, $3); }
        | IDENTIFIER BORASSIGN exp                  { $$ = createOperationAndAssignment(BOR,    false, $1, $3); }
        | IDENTIFIER BXORASSIGN exp                 { $$ = createOperationAndAssignment(BXOR,   false, $1, $3); }
        | IDENTIFIER BANDASSIGN exp                 { $$ = createOperationAndAssignment(BAND,   false, $1, $3); }
        | IDENTIFIER LSHIFTASSIGN exp               { $$ = createOperationAndAssignment(LSHIFT, false, $1, $3); }
        | IDENTIFIER RSHIFTASSIGN exp               { $$ = createOperationAndAssignment(RSHIFT, false, $1, $3); }
        | IDENTIFIER ADDASSIGN exp                  { $$ = createOperationAndAssignment(ADD,    false, $1, $3); }
        | IDENTIFIER SUBASSIGN exp                  { $$ = createOperationAndAssignment(SUB,    false, $1, $3); }
        | IDENTIFIER MULTASSIGN exp                 { $$ = createOperationAndAssignment(MULT,   false, $1, $3); }
        | IDENTIFIER DIVASSIGN exp                  { $$ = createOperationAndAssignment(DIV,    false, $1, $3); }
        | IDENTIFIER MODASSIGN exp                  { $$ = createOperationAndAssignment(MOD,    false, $1, $3); }
        | IDENTIFIER POWASSIGN exp                  { $$ = createOperationAndAssignment(POW,    false, $1, $3); }
        | IDENTIFIER vector_index ASSIGN exp        { $$ = createAssignment($1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index INC               { $$ = createOperationAndAssignment(ADD,    true,  $1, vectorIndex.pop(), new IntegerValue(1)); }
        | IDENTIFIER vector_index DEC               { $$ = createOperationAndAssignment(SUB,    true,  $1, vectorIndex.pop(), new IntegerValue(1)); }
        | INC IDENTIFIER vector_index               { $$ = createOperationAndAssignment(ADD,    false, $2, vectorIndex.pop(), new IntegerValue(1)); }
        | DEC IDENTIFIER vector_index               { $$ = createOperationAndAssignment(SUB,    false, $2, vectorIndex.pop(), new IntegerValue(1)); }
        | IDENTIFIER vector_index ORASSIGN exp      { $$ = createOperationAndAssignment(OR,     false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index ANDASSIGN exp     { $$ = createOperationAndAssignment(AND,    false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index BORASSIGN exp     { $$ = createOperationAndAssignment(BOR,    false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index BXORASSIGN exp    { $$ = createOperationAndAssignment(BXOR,   false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index BANDASSIGN exp    { $$ = createOperationAndAssignment(BAND,   false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index LSHIFTASSIGN exp  { $$ = createOperationAndAssignment(LSHIFT, false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index RSHIFTASSIGN exp  { $$ = createOperationAndAssignment(RSHIFT, false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index ADDASSIGN exp     { $$ = createOperationAndAssignment(ADD,    false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index SUBASSIGN exp     { $$ = createOperationAndAssignment(SUB,    false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index MULTASSIGN exp    { $$ = createOperationAndAssignment(MULT,   false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index DIVASSIGN exp     { $$ = createOperationAndAssignment(DIV,    false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index MODASSIGN exp     { $$ = createOperationAndAssignment(MOD,    false, $1, vectorIndex.pop(), $4);  }
        | IDENTIFIER vector_index POWASSIGN exp     { $$ = createOperationAndAssignment(POW,    false, $1, vectorIndex.pop(), $4);  }
        ;

vector_index: vector_index_begin vector_index_in  { /*$$ = vectorAccess($1, vectorIndex.pop());*/ }
        ;

vector_index_begin:                             { vectorIndex.push(new Vector<ExpressionStmt>()); }
        ;

vector_index_in: BRACKET exp BRACKET_END           { vectorIndex.peek().add(exprStack.get($2)); }
        | vector_index_in BRACKET exp BRACKET_END  { vectorIndex.peek().add(exprStack.get($3)); }
        ;

%%

    final private static Runtime runtime = new Runtime();

    private static Console       console;

    private static UserOperation helpOperation;
    private static UserOperation loadJarOperation;
    private static UserOperation loadClassOperation;
    private static UserOperation clearOperation;
    private static UserOperation exitOperation;

    private final Stack<Vector<Integer>>        params         = new Stack<Vector<Integer>>();
    private final Stack<Vector<String>>         paramNames     = new Stack<Vector<String>>();
    private final Stack<Vector<Integer>>        vector         = new Stack<Vector<Integer>>();
    private final Stack<Vector<ExpressionStmt>> vectorIndex    = new Stack<Vector<ExpressionStmt>>();
    private final Stack<Stmt>                   stmtStack      = new Stack<Stmt>();
    private final LinkedList<MultiStmt>         multiStmtStack = new LinkedList<MultiStmt>(Arrays.asList(new MultiStmt[] { new MultiStmt() }));

    private CalcParserYylex  lexer        = null;
    private ExpressionStmt   compiledExpr = null;
    private Stmt             compiledCode = null;
    private Exception        exception    = null;

    Vector<Value>          valuesStack = new Vector<Value>();
    Vector<ExpressionStmt> exprStack   = new Vector<ExpressionStmt>();

    private CalcParser(String expression, Vector<Value> memory) throws IOException, ParseException
    {
        this();

        lexer = new CalcParserYylex(expression, this, memory);

        if (yyparse() == -1)
            yyerror("INTERNAL ERROR");
    }

    private CalcParser(File file) throws IOException, ParseException
    {
        this();

        lexer = new CalcParserYylex(file, this);

        if (yyparse() == -1)
            yyerror("INTERNAL ERROR");
    }

    private int yylex()
    {
        yylval = new CalcParserVal(0);

        try
        {
            return lexer.yylex();
        }
        catch (IOException e)
        {
            yyerror(e);
            return -1;
        }
    }

    void yyerror(String message)
    {
        if (exception == null)
            exception = new ParseException(message, lexer.getCurrentPosition());
    }

    void yyerror(Exception newException)
    {
        if (exception == null)
            exception = new ParseException(newException.getMessage(), lexer.getCurrentPosition());
    }

    int pushValue(Value result)
    {
        valuesStack.add(result);

        return valuesStack.size() - 1;
    }

    int pushValue(long ival)
    {
        return pushValue(new IntegerValue(ival));
    }

    int pushValue(double dval)
    {
        return pushValue(new FloatValue(dval));
    }

    int pushValue(String sval)
    {
        return pushValue(new StringValue(sval));
    }

    int pushValue(Vector<Integer> vval)
    {
        Vector<Value> vector = new Vector<Value>();

        for (int i : vval)
            vector.add(valuesStack.get(i));

        return pushValue(new MatrixValue(vector));
    }

    private int pushExpression(ExpressionStmt expression)
    {
        exprStack.add(expression);

        return exprStack.size() - 1;
    }

    private int pushStmt(Stmt expression)
    {
        stmtStack.add(expression);

        return stmtStack.size() - 1;
    }

    private int operation(int operation, int ... indexes)
    {
        try
        {
            ExpressionStmt stmt[] = new ExpressionStmt[indexes.length-1];

            for (int i=1; i<indexes.length; i++)
                stmt[i-1] = exprStack.get(indexes[i]);

            return pushExpression(new NativeExp(lexer.getCurrentPosition(), operation, exprStack.get(indexes[0]), stmt));
        }
        catch(ParseException e)
        {
            yyerror(e);
            return -1;
        }
    }

    private int functionDeclaration(String functionName, Vector<String> paramNames, int stmtIndex)
    {
        return pushStmt(new DeclarationStmt(runtime, functionName, paramNames, stmtStack.get(stmtIndex)));
    }

    private int functionInvocation(String functionName, Vector<Integer> indexes)
    {
        ExpressionStmt[] values;

        if (indexes == null)
        {
            values = new ExpressionStmt[0];
        }
        else
        {
            values = new ExpressionStmt[indexes.size()];

            for (int i=0; i<values.length; i++)
            {
                if (indexes.get(i) == -1)
                {
                    yyerror("Bad expression");
                    return -1;
                }

                values[i] = exprStack.get(indexes.get(i));
            }
        }

        return pushExpression(new InvocationExp(lexer.getCurrentPosition(), functionName, values));
    }

    private int functionInvocation(String functionName, int objectIndex, Vector<Integer> indexes)
    {
        ExpressionStmt[] values;

        if (indexes == null)
        {
            values = new ExpressionStmt[0];
        }
        else
        {
            values = new ExpressionStmt[indexes.size()];

            for (int i=0; i<values.length; i++)
            {
                if (indexes.get(i) == -1)
                {
                    yyerror("Bad expression");
                    return -1;
                }

                values[i] = exprStack.get(indexes.get(i));
            }
        }

        ExpressionStmt object = exprStack.get(objectIndex);

        return pushExpression(new InvocationExp(lexer.getCurrentPosition(), functionName, object, values));
    }

    private int readOperation(int constantValue)
    {
        try
        {
            return pushExpression(new ValueExp(lexer.getCurrentPosition(), valuesStack.get(constantValue)));
        }
        catch (ParseException e)
        {
            yyerror(e);
            return -1;
        }
    }

    private int readOperation(String identifier)
    {
        return pushExpression(new ValueExp(lexer.getCurrentPosition(), identifier, false));
    }

    private int readOperation(String identifier, Vector<ExpressionStmt> indexes)
    {
        return pushExpression(new IndexedValueExp(lexer.getCurrentPosition(), identifier, indexes));
    }

    private int createAssignment(String identifier)
    {
        return pushExpression(new ValueExp(lexer.getCurrentPosition(), identifier, true));
    }

    private int createAssignment(String identifier, int expIndex)
    {
        return pushExpression(new ValueExp(lexer.getCurrentPosition(), identifier, exprStack.get(expIndex)));
    }

    private int createOperationAndAssignment(int operation, boolean returnOldValue, String identifier, int otherOperandIndex)
    {
        return createOperationAndAssignment(operation, returnOldValue, identifier, null, exprStack.get(otherOperandIndex));
    }

    private int createOperationAndAssignment(int operation, boolean returnOldValue, String identifier, Value otherOperand)
    {
        try
        {
            return createOperationAndAssignment(operation, returnOldValue, identifier, null, new ValueExp(lexer.getCurrentPosition(), otherOperand));
        }
        catch(ParseException e)
        {
            yyerror(e);
            return -1;
        }
    }

    private int createAssignment(String identifier, Vector<ExpressionStmt> indexes, int expIndex)
    {
        return pushExpression(new IndexedValueExp(lexer.getCurrentPosition(), identifier, indexes, exprStack.get(expIndex)));
    }

    private int createOperationAndAssignment(int operation, boolean returnOldValue, String identifier, Vector<ExpressionStmt> indexes, int otherOperandIndex)
    {
        return createOperationAndAssignment(operation, returnOldValue, identifier, indexes, exprStack.get(otherOperandIndex));
    }

    private int createOperationAndAssignment(int operation, boolean returnOldValue, String identifier, Vector<ExpressionStmt> indexes, Value otherOperand)
    {
        try
        {
            return createOperationAndAssignment(operation, returnOldValue, identifier, indexes, new ValueExp(lexer.getCurrentPosition(), otherOperand));
        }
        catch(ParseException e)
        {
            yyerror(e);
            return -1;
        }
    }

    private int createOperationAndAssignment(int operation, boolean returnOldValue, String identifier, Vector<ExpressionStmt> indexes, ExpressionStmt ... arguments)
    {
        try
        {
            ExpressionStmt assignmentExp;
            ExpressionStmt getValueExp;

            if (indexes == null)
            {
                getValueExp   = new ValueExp(lexer.getCurrentPosition(), identifier, false);

                assignmentExp = new ValueExp(lexer.getCurrentPosition(), identifier, new NativeExp(lexer.getCurrentPosition(), operation, getValueExp, arguments));
            }
            else
            {
                getValueExp   = new IndexedValueExp(lexer.getCurrentPosition(), identifier, indexes);

                assignmentExp = new IndexedValueExp(lexer.getCurrentPosition(), identifier, indexes, new NativeExp(lexer.getCurrentPosition(), operation, getValueExp, arguments));
            }

            if (returnOldValue)
                return pushExpression(new MultiExp(lexer.getCurrentPosition(), 0, getValueExp, assignmentExp));
            else
                return pushExpression(assignmentExp);
        }
        catch (ParseException e)
        {
            yyerror(e);
            return -1;
        }
    }

    static public void setConsole(JFrame frame, Console console)
    {
        helpOperation      = new HelpOperation     (console);
        loadJarOperation   = new LoadJarOperation  (console);
        loadClassOperation = new LoadClassOperation(console);
        clearOperation     = new ClearOperation    (console);
        exitOperation      = new ExitOperation     (frame, console);
    }

    static public Value calculate(String expression, Vector<Value> memory)
            throws ParseException,
                   PluginException, 
                   ExecException,
                   InvocationTargetException
    {
        try
        {
            CalcParser yyparser = new CalcParser(expression, memory);

            if (yyparser.exception != null)
            {
                if (yyparser.exception instanceof ParseException)
                    throw (ParseException) yyparser.exception;
                else if (yyparser.exception instanceof PluginException)
                    throw (PluginException) yyparser.exception;
                else if (yyparser.exception instanceof InvocationTargetException)
                    throw (InvocationTargetException) yyparser.exception;
                else
                    throw new RuntimeException(yyparser.exception);
            }
            

            if (yyparser.compiledExpr != null)
                return yyparser.compiledExpr.calculate(runtime);

            if (yyparser.compiledCode != null)
                return yyparser.compiledCode.exec(runtime).getResult();

            return null;
        }
        catch(IOException e)
        {
            throw new RuntimeException(e);
        }
    }

    static public Value calculate(File file)
            throws ParseException,
                   PluginException, 
                   ExecException,
                   InvocationTargetException
    {
        try
        {
            CalcParser yyparser = new CalcParser(file);

            if (yyparser.exception != null)
            {
                if (yyparser.exception instanceof ParseException)
                    throw (ParseException) yyparser.exception;
                else if (yyparser.exception instanceof PluginException)
                    throw (PluginException) yyparser.exception;
                else if (yyparser.exception instanceof InvocationTargetException)
                    throw (InvocationTargetException) yyparser.exception;
                else
                    throw new RuntimeException(yyparser.exception);
            }
            

            if (yyparser.compiledExpr != null)
                return yyparser.compiledExpr.calculate(runtime);

            if (yyparser.compiledCode != null)
                return yyparser.compiledCode.exec(runtime).getResult();

            return null;
        }
        catch(IOException e)
        {
            throw new RuntimeException(e);
        }
    }

    static public void addPackage(String name, String manualFile, Class packageClass)
            throws InstantiationException, 
                   IllegalAccessException, 
                   NoSuchMethodException,
                   InvocationTargetException,
                   FileNotFoundException,
                   IOException
    {
        runtime.addFunction(new Package(name, manualFile, packageClass));
    }

    static public void addPackage(String name, Class packageClass)
            throws InstantiationException, 
                   IllegalAccessException, 
                   NoSuchMethodException,
                   InvocationTargetException
    {
        runtime.addFunction(new Package(name, packageClass));
    }

    static public Vector<Package> getRegisteredPackages()
    {
        return runtime.getPackageList();
    }

    static public Package getPackage(String packageName)
    {
        return runtime.getPackage(packageName);
    }

